/* $Workfile: OtxDrv.h $   $Revision: 74 $    $Modtime: 8/25/04 12:29p $  */
#ifndef OTX_DRV_H
#define OTX_DRV_H
/******************************************************************************

  Copyright (C) 1998 - 2003 Odin TeleSystems Inc., All Rights Reserved

  @doc PublicHwApi OtxDrv

  OTX Generic Hardware Driver API

  @module
  OtxDrv | This file contains prototypes for the Driver Functions

  @contents2
  OtxDrv.h |

  @index func | OtxDrv

  @contents2
  OTX Driver Level Functions - OtxDrvXXX() | Driver level functions

  @head3 OtxDrvXXX() Function Reference |

  @index func | OtxDrv

******************************************************************************/


#ifdef __cplusplus
extern "C" {
#endif

#include "OtxDrvD.h"    // OTX Driver type definitions
#include "OtxDef.h"     // OTX specific macro definitions etc.
#include "OtxDev.h"     // OTX specific device type definitions
#include "OtxType.h"    // OTX specific type definitions


/******************************************************************************

  @func
    Returns OTX_TRUE if success on any status value.

  @comm
    It is to allow a callable routine to the API in lieu of using the macros
    provided in the C header file.  If you are using a language other than C
    this routine allows you to consistently obtain this data without having
    to port an interface routine.
    
  @xref
    <f OTX_RESULT_CODE>

******************************************************************************/
OTX_EXPORT OTX_BOOL OTX_API OtxSuccess(OTX_RESULT OtxResult);


/******************************************************************************

  @func
    Returns OTX_FALSE if failure on any status value.

  @comm
    It is to allow a callable routine to the API in lieu of using the macros
    provided in the C header file.  If you are using a language other than C
    this routine allows you to consistently obtain this data without having
    to port an interface routine.
    
  @xref
    <f OTX_RESULT_CODE>

******************************************************************************/
OTX_EXPORT OTX_BOOL OTX_API OtxFailure(OTX_RESULT OtxResult);


/******************************************************************************

  @func
    Returns the portions that make up the OTX_RESULT code.

  @comm
    This routine returns the portions of the OTX_RESULT return code.  It is
    to allow a callable routine to the API in lieu of using the macros
    provided in the C header file.  If you are using a language other than C
    this routine allows you to consistently obtain this data without having
    to port an interface routine.
    
  @xref
    <f OTX_RESULT_CODE>
    <f OTX_RESULT_SEVERITY>
    <f OTX_RESULT_EX_SOURCE_ID>
    <f OTX_RESULT_EX_LINE_NUM>

******************************************************************************/
OTX_EXPORT void OTX_API OtxResultCode(
  OTX_RESULT OtxResult,       // @parm OTX Result code to process.
  OTX_UINT32 *Result,         // @parm pointer to store the Result Code in.
  OTX_UINT32 *Severity,       // @parm pointer to store the Severity Code in.
  OTX_UINT32 *SourceId,       // @parm pointer to store the SourceID in.
  OTX_UINT32 *LineNum         // @parm pointer to store the Line Number in.
);


/*******************************************************************************

  @func
    Returns the Severity Code from the OTX_RESULT code.

  @comm
    This routine returns the Severity Code from the OTX_RESULT return code.
    It is to allow a callable routine to the API in lieu of using the macros
    provided in the C header file.  If you are using a language other than C
    this routine allows you to consistently obtain this data without having
    to port an interface routine.
    
  @xref
    <f OTX_RESULT_CODE>
    <f OTX_RESULT_SEVERITY>

*******************************************************************************/
OTX_EXPORT OTX_UINT32 OTX_API OtxResultSeverity(OTX_RESULT OtxResult);


/*******************************************************************************

  @func
    Returns the Source ID from the OTX_RESULT code.

  @comm
    This routine returns the Source ID from the OTX_RESULT return code.
    It is to allow a callable routine to the API in lieu of using the macros
    provided in the C header file.  If you are using a language other than C
    this routine allows you to consistently obtain this data without having
    to port an interface routine.
    
  @xref
    <f OTX_RESULT_CODE>
    <f OTX_RESULT_EX_SOURCE_ID>

*******************************************************************************/
OTX_EXPORT OTX_UINT32 OTX_API OtxResultSourceId(OTX_RESULT OtxResult);


/*******************************************************************************

  @func
    Returns the Line Number from the OTX_RESULT code.

  @comm
    This routine returns the Line Number from the OTX_RESULT return code.
    It is to allow a callable routine to the API in lieu of using the macros
    provided in the C header file.  If you are using a language other than C
    this routine allows you to consistently obtain this data without having
    to port an interface routine.
    
  @xref
    <f OTX_RESULT_CODE>
    <f OTX_RESULT_EX_LINE_NUM>

*******************************************************************************/
OTX_EXPORT OTX_UINT32 OTX_API OtxResultLineNum(OTX_RESULT OtxResult);


/*******************************************************************************

  @func
    Returns the Error Code from the OTX_RESULT code.

  @comm
    This routine returns the Error Code from the OTX_RESULT return code.
    It is to allow a callable routine to the API in lieu of using the macros
    provided in the C header file.  If you are using a language other than C
    this routine allows you to consistently obtain this data without having
    to port an interface routine.
    
  @xref
    <f OTX_RESULT_CODE>

*******************************************************************************/
OTX_EXPORT OTX_UINT32 OTX_API OtxResultErrorCode(OTX_RESULT OtxResult);


/******************************************************************************

  @func
    Opens a communication channel to the OTX Network Driver API library.

  @comm
    This is the first function to be called in OTX board application, it sets 
    up a communication channel with the driver.

    Its complement is OtxDrvReleaseLib which closes the communication channel 
    and is the last function to be called.
    
  @ex |
    // start of OTX applicaton
    result = OtxNetConnectLib();
    ......
    ......
    // end of OTX application
    result = OtxDrvReleaseLib();  

  @xref
    <f OtxDrvReleaseLib>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxNetConnectLib(
  void
);


/******************************************************************************

  @func
    Opens a communication channel to the driver API library.

  @comm
    This is the first function to be called in OTX board application, it sets 
    up a communication channel with the driver.

    Its complement is OtxDrvReleaseLib which closes the communication channel 
    and is the last function to be called.
    
  @ex |
    // start of OTX applicaton
    result = OtxDrvConnectLib();
    ......
    ......
    // end of OTX application
    result = OtxDrvReleaseLib();  

  @xref
    <f OtxDrvReleaseLib>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvConnectLib(
  void
);


/******************************************************************************

  @func
    Releases all references that this driver is holding.

  @comm
    This is the last function to be called in OTX board application. 
    All handles that an application is holding are closed automatically when a
    process terminates. If the application is about to terminate, calling
    this function is not necessary. However a server application may need
    this function if it never terminates.

    see example on
    @ex <f OtxDrvConnectLib> |
  
  @xref
    <f OtxDrvConnectLib>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvReleaseLib(
  void
);


/******************************************************************************

  @func
    Gets the current time from the host.

  @ex |     
    result = OtxDrvGetCurrentTime(&nTimeStamp); 

  @xref
    <t OTX_DATETIME>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvGetCurrentTime(
  OUT OTX_DATETIME  *pnTimeStamp        // @parm Current time of the host
);


/******************************************************************************

  @func
    Get the an attribute type of a device

  @comm
    This function can, for example, be used to retrieve the revision of the
    currently loaded driver, or a firmware revision of a physical device.
    Different devices have different attributes, see the header file for the
    device in question for specific information

  @ex |
    
    // Read Serial Number from the board
    OtxAttribTypeE  eAttribType;    
    OTX_UINT32      cbAttribValue;  
    OTX_BOOL        bReadOnly;      
    char *pszSerialNumber;
    
    result = OtxDrvGetAttributeType(
                g_hPhysDevs.hBoardController, 
                OTX_ATTR_BRD_SERIAL_NUMBER, 
                &eAttribType, 
                &cbAttribValue, 
                &bReadOnly); 

    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);

    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    
    result = OtxDrvGetAttributeValueSz(
                g_hPhysDevs.hBoardController, 
                OTX_ATTR_BRD_SERIAL_NUMBER, 
                pszSerialNumber, 
                cbAttribValue); 

    // display 
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    
    free(pszSerialNumber);

  
  @xref
    <f OtxDrvGetAttributeValue>
    <nl><f OtxDrvSetAttributeValue>
    <nl><f OtxDrvResultCode2String>
    <nl><f OtxDrvEventCode2String>
    <nl><f OtxDrvConnectLib>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvGetAttributeType(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device being queried
  IN  OTX_ATTRIB_ID   nAttribId,        // @parm ID of attribute
  OUT OtxAttribTypeE  *peAttribType,    // @parm Type of the attribute, if attribute exists
  OUT OTX_UINT32      *pcbAttribValue,  // @parm Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
  OUT OTX_BOOL        *pbReadOnly       // @parm Is attribute writable or read-only

);


/******************************************************************************

  @func
    Get an attribute (Boolean type) of a device 

  @ex |
    
    // Query Library revision  
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(
                OTX_LIB_HANDLE, 
                OTX_ATTR_LIB_REV_STR, 
                szVersionInfo, 
                255); 

    result = OtxDrvGetAttributeValueBool(
                OTX_LIB_HANDLE, 
                OTX_ATTR_LIB_TYPE_STATIC, 
                &bIsStaticLinkedLibrary);
                
    result = OtxDrvGetAttributeValueSz(
                OTX_LIB_HANDLE, 
                OTX_ATTR_LIB_FILE_NAME, 
                szLibFileName, 
                64); 
    
    // display 
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  

  @xref
    <f OtxDrvSetAttributeValueBool>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvGetAttributeValueBool(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device being queried
  IN  OTX_ATTRIB_ID   nAttribId,        // @parm ID of attribute
  OUT OTX_BOOL        *pAttribValue     // @parm Buffer to hold the value of the attribute
);


/******************************************************************************

  @func
    Get an attribute (unsigned long type) of a device

  @ex |
    
    // get board Device ID and SubSystem ID
    OTX_UINT32 nDeviceId;
    OTX_UINT32 nSubsystemId;

    result = OtxDrvGetAttributeValueUint32(
                g_hPhysDevs.hBoardController, 
                OTX_ATTR_BRD_DEVICE_ID, 
                &nDeviceId); 
    
    result = OtxDrvGetAttributeValueUint32(
                g_hPhysDevs.hBoardController, 
                OTX_ATTR_BRD_SUBSYSTEM_ID, 
                &nSubsystemId); 
    
    // display
    DemoLogf("Board Device ID: %d, SubsystemId: 0x%.4X\n", nDeviceId, nSubsystemId);

  @xref
    <f OtxDrvSetAttributeValueUint32>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvGetAttributeValueUint32(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device being queried
  IN  OTX_ATTRIB_ID   nAttribId,        // @parm ID of attribute
  OUT OTX_UINT32      *pAttribValue     // @parm Buffer to hold the value of the attribute
);


/******************************************************************************

  @func
    Get an attribute (char* type) of a device

  @comm
    see example on <f OtxDrvGetAttributeType>
   
  @xref
    <f OtxDrvSetAttributeValueSz>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvGetAttributeValueSz(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device being queried
  IN  OTX_ATTRIB_ID   nAttribId,        // @parm ID of attribute
  OUT char            *pAttribValueBuf, // @parm Buffer to hold the value of the attribute
  IN  OTX_INT32       cbAttribValueBuf  // @parm Size of the attribValueBuf in bytes
);


/******************************************************************************

  @func
    Get an attribute (void* type) of a device

  @comm
    This function can, for example be used to retrieve the revision of the
    currently loaded driver, or a firmware revision of a physical device.

  @ex |

    // get driver version
    OtxVersionInfo sVersionInfo;
    char szVersionInfo[255];

    OtxDrvGetAttributeValueSz(
       0, 
       OTX_ATTR_DRV_REV_STR, 
       szVersionInfo, 
       255);
    OtxDrvGetAttributeValue(
       0, 
       OTX_ATTR_DRV_REV_NUM, 
       &sVersionInfo, 
       sizeof(OtxVersionInfo), 
       NULL);
    
    // display
    DemoLogf("Driver version string     %s\n", szVersionInfo);
    DemoLogf("Driver major version      %d\n", sVersionInfo.nMjRev);
    DemoLogf("Driver minor version      %d\n", sVersionInfo.nMnRev);
    DemoLogf("Driver status             %s\n", sVersionInfo.bRelStatus ? "Released" : "Preliminary");
    DemoLogf("Driver preliminary level  %d\n", sVersionInfo.nPrelLevel);

  
  @xref
    <f OtxDrvGetAttributeType>
    <nl><f OtxDrvSetAttributeValue>
    <nl><f OtxDrvResultCode2String>
    <nl><f OtxDrvEventCode2String>
    <nl><f OtxDrvConnectLib>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvGetAttributeValue(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device being queried
  IN  OTX_ATTRIB_ID   nAttribId,        // @parm ID of attribute
  OUT OTX_PVOID       pAttribValueBuf,  // @parm Buffer to hold the value of the attribute
  IN  OTX_UINT32      cbAttribValueBuf, // @parm Size of the attribValueBuf in bytes
  OUT OTX_UINT32      *pcbAttribValue   // @parm Bytes actually used in attribValueBuf
);


/******************************************************************************

  @func
    Set an attribute (Boolean type) of a device

  @ex |
    // toggle OTX_ATTR_T1E1_AUTOMATIC_RESET attribute
    OTX_BOOL bAutomaticResetMode;

    result = OtxDrvGetAttributeValueBool(
                g_hPhysDevs.hLi[nCurrentLi], 
                OTX_ATTR_T1E1_AUTOMATIC_RESET, 
                &bAutomaticResetMode); 

    bAutomaticResetMode = !bAutomaticResetMode;
      
    result = OtxDrvSetAttributeValueBool(
                g_hPhysDevs.hLi[nCurrentLi], 
                OTX_ATTR_T1E1_AUTOMATIC_RESET, 
                bAutomaticResetMode); 

  @xref
    <f OtxDrvGetAttributeValueBool>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvSetAttributeValueBool(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device being queried
  IN  OTX_ATTRIB_ID   nAttribId,        // @parm ID of attribute
  IN  OTX_BOOL        attribValue       // @parm Value of the attribute, if attribute exists
);


/******************************************************************************

  @func
    Set an attribute (unsigned long type) of a device

  @ex | 
    // set Receive Input Threshold for current Li
    char strBuf[12];
    OTX_UINT32 nThresholdInMilliVolts;

    result = OtxDrvGetAttributeValueUint32(
                g_hPhysDevs.hLi[nCurrentLi], 
                OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, 
                &nThresholdInMilliVolts); 
    
    // display 
    DemoLogf("Threshold (%d mV): ", nThresholdInMilliVolts);

    nThresholdInMilliVolts = 100;

    // set attribute  
    result = OtxDrvSetAttributeValueUint32(
                g_hPhysDevs.hLi[nCurrentLi], 
                OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, 
                nThresholdInMilliVolts); 
    
    // get attribute 
    result = OtxDrvGetAttributeValueUint32(
                g_hPhysDevs.hLi[nCurrentLi], 
                OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, 
                &nThresholdInMilliVolts); 
        
    // display
    DemoLogf("Receive Input Threshold is changed to %d mV.\n", nThresholdInMilliVolts);
      
  @xref
    <f OtxDrvGetAttributeValueUint32>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvSetAttributeValueUint32(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device being queried
  IN  OTX_ATTRIB_ID   nAttribId,        // @parm ID of attribute
  IN  OTX_UINT32      attribValue       // @parm Value of the attribute, if attribute exists
);


/******************************************************************************

  @func
    Set an attribute (char* type) of a device

  @ex |
    // configuring Thor-2-PCMCIA series board for dongle type 
    result = OtxDrvSetAttributeValueSz(
                g_hPhysDevs.hBoardController, 
                OTX_ATTR_BRD_PCMCIA_DONGLE_TYPE, 
                "HMA-1079-2-1.0"); 


  @xref
    <f OtxDrvGetAttributeValueSz>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvSetAttributeValueSz(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device being queried
  IN  OTX_ATTRIB_ID   nAttribId,        // @parm ID of attribute
  IN  char            *pAttribValue     // @parm Value of the attribute, if attribute exists
);


/******************************************************************************

  @func
    Set an attribute of a device

  @comm
    This function can, for example be used to disable and enable certain
    functions of a device, for example, enable or disable reception
    of Caller ID for a Pots device . Note that some attributes are read-only. 
    If an attempt to set a read-only attribute is made, an error code is 
    returned as result.

  @xref
    <f OtxDrvGetAttributeType>
    <nl><f OtxDrvGetAttributeValue>
    <nl><f OtxDrvResultCode2String>
    <nl><f OtxDrvEventCode2String>
    <nl><f OtxDrvConnectLib>

  @todo

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvSetAttributeValue(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device being queried
  IN  OTX_ATTRIB_ID   nAttribId,        // @parm ID of attribute
  OUT OTX_UINT8       attribValueBuf[], // @parm Value of the attribute, if attribute exists
  IN  OTX_UINT32      cbAttribValueBuf  // @parm Size in bytes of the buffer for the value
);


/******************************************************************************

  @func
    Gets a handle to an already existing Device Object

  @comm

    The returned handle can only be used for query operations such as
    getting attributes and for opening child devices which requires the
    handle of the parent.

  @ex |
      
    // Get a handle to the ASM Daughter Board
    result = OtxDrvGetDeviceHandle(
                g_hPhysDevs.hThorController, 
                OTX_DEVICE_VIDAR_5x_ASM, 
                0, 
                &g_hPhysDevs.hVidarAsmController); 

  @xref
    <f OtxDrvOpenPhysicalDevice>
    <nl><t OtxPhysicalDeviceTypeE>
    <nl><t OtxLogicalDeviceTypeE>
    <nl><t OTX_HANDLE>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvGetDeviceHandle(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device being queried
  IN  OtxDeviceType   nDeviceType,      // @parm Device type, either OtxPhysicalDeviceTypeE or OtxLogicalDeviceTypeE
  IN  OTX_INT32       nDevice,          // @parm Sequence number of device within it's class
  OUT OTX_HANDLE      *phDevice         // @parm Buffer to hold returned device handle
);


/******************************************************************************

  @func
    Opens a Device Object representing a physical hardware device, which can
    be an entire board or an integrated circuit (chip) on a board. The number 
    of physical devices is set in hardware and cannot change dynamically.

  @comm
    Each physical devices in the system are identified by 3 parameters
    - Parent          : Physical Device contaning the device
    - Type            : Type of the Physical Device
    - Sequence Number : Number of the device within the host and with its type

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the placeholder and controller
    of other physical devices on the board.

  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions which then
    can be represented by a multiple of logical devices.

    The chip devices can be further categorized as follows:

  - Line Interface Devices: A hardware device that interfaces with the external
    network and converts/maps the non-TTL level external data interface to the
    internal serial data highways. Typically an actual line interface transceiver
    chip on a board. Examples of Line Interface Devices include T1/E1, BRI, POTS,
    T3/E3 Lis.

  - Processor Devices: A processor (general purpose micro processor or Digital
    Signal Processor) is an integrated circuit that implements certain functionality
    (logical devices) under software control.

  - Switch Device: A device connecting and disconnecting data-paths.

  - Other miscellaneous chips.

  @ex |

    // open mother board 
    result = OtxDrvOpenPhysicalDevice(
                0,                      // '0' for mother board
                OTX_DEVICE_THOR_PCI, 
                nboardNo, 
                nboardNo, 
                g_hEventQueue, 
                &g_hPhysDevs.hBoardController); 

    
    // open daughter board or chip devices
   
    // open two T1/E1 Line Interfaces
    for (i = 0; i < 2; i++) {
      result = OtxDrvOpenPhysicalDevice(
                  g_hPhysDevs.hBoardController, 
                  OTX_DEVICE_LI_T1E1, 
                  i, 
                  i, 
                  g_hEventQueue, 
                  &g_hPhysDevs.hLi[i]);
    }


  @xref
    <f OtxDrvClose>
    <nl><t OtxPhysicalDeviceTypeE>
    <nl><t OtxLogicalDeviceTypeE>
    <nl><t OTX_HANDLE>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvOpenPhysicalDevice(
  IN  OTX_HANDLE      hHostDevice,      // @parm Parent (typically Board Device) who owns the device we are connecting to
                                        //       Specify NULL when opening a board device
  IN  OtxPhysicalDeviceTypeE eDevType,  // @parm Class of device we are looking for. See <t OtxPhysicalDeviceTypeE>
  IN  OTX_INT32       nDevice,          // @parm Sequence number of device within it's class
  IN  OTX_INT32       nEventSourceId,   // @parm User specified ID to be used by this device as the Source ID for the events generated by the device
  IN  OTX_HANDLE      hEventQueue,      // @parm Handle to separate event queue (specify OTX_INVALID_HANDLE_VALUE to create device specific event queue)
  OUT OTX_HANDLE      *phPhysDevice     // @parm Handle to the new physical device (if result code OTX_S_OK)
);


/******************************************************************************

  @func
    Opens a Device Object representing a physical hardware device, for which 
    a handle has already been acquired (e.g with <f OtxDrvEnumDevice>). Otherwise
    this function has the same functionality as <f OtxDrvOpenPhysicalDevice>.

  @comm
    see example on <f OtxDrvEnumDevice> 

  @xref
    <f OtxDrvEnumDevice>
    <f OtxDrvOpenPhysicalDevice>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvOpenPhysicalDeviceEx(
  IN  OTX_HANDLE      hPhysDevice,      // @parm Handle to the physical device
  IN  OTX_INT32       nEventSourceId,   // @parm User specified ID to be used by this device as the Source ID for the events generated by the device
  IN  OTX_HANDLE      hEventQueue       // @parm Handle to separate event queue (specify OTX_INVALID_HANDLE_VALUE to create device specific event queue)
);                                      


/******************************************************************************

  @func
    Create a Device Object to represent a logical device which is implemented
    with hardware, firmware or software within a physical device.

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not pre-determined. More logical devices can
    be requested by the user and allocated by the system if needed. Examples of
    logical devices include HDLC Receiver, HDLC Sender, Tone Detector, 
    Tone Generator, etc.

    The following Logical Devices are supported by the OTX Driver:

    - Tone Devices  : Logical devices for tone generation and reception (See OtxTone.h)
    - Data Devices  : Devices for sending, receiving, and manipulation of
                      Raw data (See OtxData.h)
    - HDLC Devices  : Logical devices implementing HDLC Sender and Receiver
                      Devices (See OtxHdlc.h)
    - CAS Devices   : Devices implementin Channel Associated Signalling (CAS)
                      functions (See OtxCas1.h)
    - Modem Devices : Devices implementing Modem emulation
    - Fax Devices   : Devices implementin Fax emulation
    - Voice Devices : Devices implementing Voice Codecs, echo cancellation,
                      silence supression, etc.

  @ex |
    
    // create Hdlc Senders and Receivers logical devices hosted in the physical Line Interface devices
    for (i = 0; i < 2; i++) {
      result = OtxDrvCreateLogicalDevice( 
                  g_hPhysDevs.hLi[i], 
                  OTX_LDEVICE_HDLC_SENDER, 
                  g_hEventQueue, 
                  i, 
                  &g_hLogicDevs.hHdlcSender[i]); 
      result = OtxDrvCreateLogicalDevice( 
                  g_hPhysDevs.hLi[i], 
                  OTX_LDEVICE_HDLC_RECEIVER, 
                  g_hEventQueue, 
                  i, 
                  &g_hLogicDevs.hHdlcReceiver[i]); 
    }


  @xref
    <f OtxDrvClose>
    <nl><t OtxLogicalDeviceTypeE>
    <nl><t OtxPhysicalDeviceTypeE>
    <nl><t OTX_HANDLE>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvCreateLogicalDevice(
  IN  OTX_HANDLE      hHostDevice,      // @parm Parent (typically a physical device) who owns the logical device we are connecting to
  IN  OtxLogicalDeviceTypeE eDevType,   // @parm Class of device we are looking for, as known by the module in question. See also <t OtxLogicalDeviceTypeE>
  IN  OTX_HANDLE      hEventQueue,      // @parm Handle to separate event queue (0 creates device specific event queue)
  IN  OTX_INT32       nEventSourceId,   // @parm User specified ID to be used by this device as the Source ID for the events generated by the device
  OUT OTX_HANDLE      *phLogicDevice    // @parm Handle to the new logical device (if result code OTX_S_OK)
);


/******************************************************************************

  @func
    Enables a physical or logical driver device. Enabling a device makes the
    object to accept I/O requests and enables event notification if so
    configured.

  @ex |
    // open mother board and enable it
    result = OtxDrvOpenPhysicalDevice(
                0,                      // '0' for mother board
                OTX_DEVICE_THOR_PCI, 
                nboardNo, 
                nboardNo, 
                g_hEventQueue, 
                &g_hPhysDevs.hBoardController); 

   result = OtxDrvEnable(
                g_hPhysDevs.hBoardController,  
                OTX_TASK_SYNC, 
                5000L); 

  @xref
    <f OtxDrvDisable>
    <nl><f OtxDrvReset>
    <nl><f OtxDrvCancelTask>
    <nl><t OTX_HANDLE>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvEnable(
  IN OTX_HANDLE       hObject,          // @parm Any <t OTX_HANDLE>
  IN OTX_TASK_REF     nTaskRef,         // @parm See <t OTX_TASK_REF>
  IN OTX_TIME         nWaitMaxMs        // @parm See <t OTX_TIME>
);


/******************************************************************************

  @func
    Disables a driver object. Disabling an object makes the object reject
    all IO requests. Only certain control commands can be issued to a disabled
    object (such as configuration and reset commands). A disabled object can not
    generate events. Events still in the event queue are not removed by this
    command.

  @ex |
    // disable board interrupts
    OtxDrvDisable(
       g_hPhysDevs.hBoardController, 
       1L, 
       OTX_TIME_INFINITY);

  @xref
    <f OtxDrvEnable>
    <nl><f OtxDrvReset>
    <nl><f OtxDrvCancelTask>
    <nl><t OTX_HANDLE>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvDisable(
  IN OTX_HANDLE       hObject,          // @parm Any <t OTX_HANDLE>
  IN OTX_TASK_REF     nTaskRef,         // @parm See <t OTX_TASK_REF>
  IN OTX_TIME         nWaitMaxMs        // @parm See <t OTX_TIME>
);


// @todo Not all devices maintain a public state yet
/******************************************************************************

  @func
    Gets the status of a device.

  @comm
    Any device can be in one of the 3 states:
    "WORKING," "MANUALLY BLOCKED," or "AUTO BLOCKED."  Disabling device with 
    function <f OtxDrvDisable> forces the device to a state MANUALLY BLOCKED. 
    Enabling of a device with function <f OtxDrvEnable> attempts to set the 
    device into a WORKING state. However, if the device does not operate 
    correctly, e.g., it generates excessive amounts of interrupts, the driver 
    can automatically block the device and force it to state "AUTO BLOCKED."

  @ex |
      
    // toggle OTX_DEVICE_WORKING on device
    OTX_RESULT result;
    OtxDeviceStateE  eDevState;
    OTX_HANDLE hHandle = g_hLogicDevs.hDtmfDialler[nCurrentLi];
      
    result = OtxDrvGetState(
                hHandle, 
                &eDevState); 
   
    if (eDevState == OTX_DEVICE_WORKING) {
      // disable sending device on current Li
      result = OtxDrvDisable(
                  hHandle,  
                  OTX_TASK_SYNC, 
                  5000L); 
    } else {
      // enable sending device on current Li
      result = OtxDrvEnable(
                  hHandle,  
                  OTX_TASK_SYNC, 
                  5000L); 
    }
 

  @xref
    <f OtxDrvEnable>
    <nl><f OtxDrvDisable>
    <nl><t OtxDeviceStateE>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvGetState(
  IN OTX_HANDLE       hObject,          // @parm Handle to a driver object to get state info about
  OUT OtxDeviceStateE *peDevState       // @parm State of the device. See also <t OtxDeviceStateE>
);


/******************************************************************************

  @func
    Resets a device. It is not normally necessary to reset a device.
    A reset is normally done to resolve and error condition. Some objects must 
    be disabled before they can be reset.

    Events in the event queue for the particular object are removed by this
    command.

  @ex |
    // reset HdlcReceiver on current Li
    result = OtxDrvReset(g_hLogicDevs.hHdlcReceiver[nCurrentLi]); 

  @xref
    <f OtxDrvEnable>
    <nl><f OtxDrvDisable>
    <nl><t OTX_HANDLE>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvReset(
  IN OTX_HANDLE       hObject           // @parm Handle to a object to reset
);


/******************************************************************************

  @func
    Close a physical or logical device or event queue or pipe.

  @comm
    If the device is being used, it may not be possible to close it.
    E.g. if devices are connected to an event queue, to close the event queue 
    may not succeed.

  @ex |
    // close Hdlc Receivers logical devices hosted in the physical Line Interface devices
    for (i = 0; i < 2; i++) {
      if (g_hLogicDevs.hHdlcReceiver[i]) {
        result = OtxDrvClose(g_hLogicDevs.hHdlcReceiver[i]); 
      }
    }

  @xref
    <f OtxDrvOpenPhysicalDevice>
    <nl><f OtxDrvCreateLogicalDevice>
    <nl><t OTX_HANDLE>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvClose(
  IN OTX_HANDLE       hObject           // @parm Object handle to a physical/logical device, pipe or event queue
);


/******************************************************************************

  @func
    Create a logical pipe. (One direction, one end is input, other is output)
    A pipe is always needed to carry data between two physical devices or 
    between a physical device and a logical device.

  @comm
    A Pipe is a logical entity representing a connection between two or more 
    devices. A Pipe has a direction, thus two Pipes are needed for a duplex 
    connection. A pipe can connect any two devices regardless of where these 
    two devices are located. The devices can be either physical or logical or 
    both.

    A pipe can be created for a data rate between 8 kbit/s and 2.048Mbit/s
    (2.048 Mbit/s max is standard. Highways can be operated at up to 8.192Mbits/s).
    A pipe is a configured entity and it is overlaid on internal highways.
    A pipe can constitute a single time-slot on a Highway or it can be a 
    superchannel of any number of time slots. The pipe is mapped on the 
    physical highways when it is connected to a device with function 
    <f OtxDrvConnectPipe>

    see example on <f OtxDrvConnectPipe>

  @xref
    <f OtxDrvConnectPipe>
    <nl><f OtxDrvClose>
    <nl><f OtxDrvDisconnectPipe>
    <nl><t OTX_HANDLE>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvCreatePipe(
  IN  OTX_INT32       nCapacity,        // @parm Capacity in kbps (only 8kbps increments allowed)
  OUT OTX_HANDLE      *phPipe           // @parm Handle to the new pipe device (if result code OTX_S_OK)
);


/******************************************************************************

  @func
    Connects one end of a Pipe to a Physical Device.

  @comm
    A pipe is always used to connect two physical devices and it is always forms
    a one-to-one data path between two physical devices. I.e. every pipe must
    have exactly one physical device connected to its input and exactly one 
    physical device connected to its output.

    Multiple logical devices (implemented within a physical device) can be
    connected within a physical device to use data from a pipe output. To 
    connect logical devices to each other and to a pipe, use function
    <f OtxDrvConnectLogicalDevice>.

    To specify which timeslot (or timeslots) the pipe will be connected to, 
    a connection mask can be used. Each bit in the connection mask corresponds
    to 8 kpbs capacity. A 64 kpbs pipe corresponds to 8 bits (1 octet) in the 
    connection mask.

    The connection mask is only needed for devices where an exact mapping
    of the pipe onto a physical interface (i.e. specific time-slots) is needed,
    e.g. in the case of a multichannel line interface where a specific timeslot 
    is used for signalling. It is important that the mapping matches the 
    capacity of the pipe exactly. If the time-slots to be used do not matter, 
    or if they cannot be changed by the user, specify coMaskSize = 0 and 
    sqConnectMask = NULL.

    The connection mask is also needed for pipes of a capacity <lt> 64 kpbs. 
    E.g. if a 56kpbs pipe is specified for an HDLC device, the timeslot mask 
    for that pipe must be specified to 0xFE. Please see example below:

  @ex 1 56kpbs pipe connected to TS1 |
      OTX_UINT8 sqTimeSlotMask[32];
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[1] = 0xFE;  // TS1 is a 56kpbs pipe

  @ex 2 64kpbs pipe connected to TS16 |
      OTX_UINT8 sqTimeSlotMask[32];
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[16] = 0xFF;  // TS16 is a 64kpbs pipe

  @ex 3 128kpbs pipe connected to TS5 and TS7 |
      OTX_UINT8 sqTimeSlotMask[32];
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[5] = 0xFF;  
      sqTimeSlotMask[7] = 0xFF;  

  @ex | 
    //connect the Logical Device in the Line Interface (one for each direction)
  
    for (i = 0; i < 2; i++) {
      ts = 5; 
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      sqTimeSlotMask[ts] = 0xFF;
      nCapacity = 64;
      
      // Create an incoming and outgoing Pipe representing the external T1/E1
      result = OtxDrvCreatePipe(
                 nCapacity, 
                 &g_sqLiHdlcReceiver[i].hPipe); 
      result = OtxDrvCreatePipe(
                 nCapacity, 
                 &g_sqLiHdlcSender[i].hPipe); 
      
      // connect external T1/E1 pipes. Only connect ONE end of each pipe. 
      // External Interface is LI Highway Number 1
      result = OtxDrvConnectPipe(
                 g_sqLiHdlcReceiver[i].hPipe,  
                 OTX_PIPE_OUTPUT, 
                 g_hPhysDevs.hLi[i],
                 1, 
                 sqTimeSlotMask,
                 32); 
      result = OtxDrvConnectPipe(
                 g_sqLiHdlcSender[i].hPipe, 
                 OTX_PIPE_INPUT,  
                 g_hPhysDevs.hLi[i],
                 1, 
                 sqTimeSlotMask,
                 32); 

      // connect the logical devices within the Li
      result = OtxDrvConnectLogicalDevice(
                 g_sqLiHdlcSender[i].hHandle, 
                 -1, 
                 g_sqLiHdlcSender[i].hPipe, 
                 -1); 
      result = OtxDrvConnectLogicalDevice(
                 g_sqLiHdlcReceiver[i].hPipe,           
                 -1, 
                 g_sqLiHdlcReceiver[i].hHandle, 
                 -1); 
    }

  @ex | 
    //connect the Logical Device in DSP, connect BOTH ends of each pipe
    
    for (i = 0; i < 2; i++) {

      // Receiving direction
      {
        ts = 1;
        memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
        sqTimeSlotMask[ts] = 0xFF;
        nCapacity = 64;
        
        result = OtxDrvCreatePipe(
                   nCapacity, 
                   &g_sqDspHdlcReceiver[i].hPipe); 
        
        result = OtxDrvConnectPipe(
                   g_sqDspHdlcReceiver[i].hPipe, 
                   OTX_PIPE_OUTPUT,  
                   g_hPhysDevs.hDsp[i], 
                   -1, 
                   sqTimeSlotMask, 
                   32); 
        result = OtxDrvConnectPipe(
                   g_sqDspHdlcReceiver[i].hPipe, 
                   OTX_PIPE_INPUT,   
                   g_hPhysDevs.hLi[i],  
                   -1, 
                   sqTimeSlotMask, 
                   32); 
        
        result = OtxDrvConnectLogicalDevice(
                   g_sqDspHdlcReceiver[i].hPipe,  
                   -1,  
                   g_sqDspHdlcReceiver[i].hHandle,    
                   -1); 
      }
                
      // Sending direction     
      {
        ts = 1;
        memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
        sqTimeSlotMask[ts] = 0xFF;
        nCapacity = 64;
        
        result = OtxDrvCreatePipe(
                   nCapacity, 
                   &g_sqDspHdlcSender[i].hPipe); 
        
        result = OtxDrvConnectPipe(
                   g_sqDspHdlcSender[i].hPipe, 
                   OTX_PIPE_INPUT,  
                   g_hPhysDevs.hDsp[i], 
                   -1, 
                   sqTimeSlotMask, 
                   32); 
        result = OtxDrvConnectPipe(
                   g_sqDspHdlcSender[i].hPipe, 
                   OTX_PIPE_OUTPUT,   
                   g_hPhysDevs.hLi[i],  
                   -1, 
                   sqTimeSlotMask, 
                   32); 
        
        result = OtxDrvConnectLogicalDevice(
                   g_sqDspHdlcSender[i].hHandle,  
                   -1, 
                   g_sqDspHdlcSender[i].hPipe,  
                   -1); 
      }
   }

  @devnote
    This function can only used for Physical devices. Logical Devices have
    to be connected with the <f OtxDrvConnectLogicalDevice> function.

  @xref
    <f OtxDrvCreatePipe>
    <nl><f OtxDrvClose>
    <nl><f OtxDrvDisconnectPipe>
    <nl><f OtxDrvOpenPhysicalDevice>
    <nl><f OtxDrvConnectLogicalDevice>
    <nl><t OtxPipeEndE>
    <nl><t OTX_HANDLE>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvConnectPipe(
  IN  OTX_HANDLE      hPipe,            // @parm Pipe to be connected to a Physical Device.
  IN  OtxPipeEndE     ePipeEnd,         // @parm End of the pipe (input vs. output) to be connected. See Also <t OtxPipeEndE>
  IN  OTX_HANDLE      hDevice,          // @parm Physical Device to be connected to
  IN  OTX_INT32       nHighway,         // @parm Physical Highway to be used. -1 = don't care
  IN  OTX_UINT8       sqConnectMask[],  // @parm Connection mask for mapping the Pipe Bits to time-slots and bits
                                        //       on the PCM highway connected to the physical device
                                        //       Can be NULL if don't care and coMaskSize = 0
                                        //       Each octet in the mask represents a timeslot on the highway
                                        //       A connection mask is needed for pipes connected on OTX PCMCIA card
                                        //       and also for pipes with a capacity <lt> 64 kpbs.
                                        //       Please note that the capacity specified in the connection
                                        //       mask must match the capacity used when the pipe was created with <f OtxDrvCreatePipe>.
  IN  OTX_UINT32      coMaskSize        // @parm Size of connection mask for the highway connection (in octets). 0 = don't care.
);


/******************************************************************************

  @func
    Disconnects one end of a Pipe from a Physical Device.

  @comm
    After disconnection, the pipe can be re-connected to another physical
    device with function <f OtxDrvConnectPipe>

  @ex |
    OTX_INT32 ts; 
    OTX_UINT8 sqTimeSlotMask[32];

    // disonnect pipe
    result = OtxDrvDisconnectPipe(
                g_sqDtmfDialer[1].hPipe, 
                OTX_PIPE_INPUT); 
    result = OtxDrvDisconnectPipe(
                g_sqDtmfDialer[2].hPipe, 
                OTX_PIPE_INPUT); 

    // reconnect pipes, arbitrary timeslot on the DSP highway
    result = OtxDrvConnectPipe(
                g_sqDtmfDialer[2].hPipe, 
                OTX_PIPE_INPUT,  
                g_hPhysDevs.hDsp[0], 
                -1, 
                NULL, 
                0); 
    result = OtxDrvConnectPipe(
                g_sqDtmfDialer[1].hPipe, 
                OTX_PIPE_INPUT,  
                g_hPhysDevs.hDsp[0], 
                -1, 
                NULL, 
                0); 

    // reconnect logical devices
    result = OtxDrvConnectLogicalDevice(
                g_sqDtmfDialer[1].hHandle, 
                -1, 
                g_sqDtmfDialer[1].hPipe,   
                -1); 
    result = OtxDrvConnectLogicalDevice(
                g_sqDtmfDialer[2].hHandle, 
                -1, 
                g_sqDtmfDialer[2].hPipe,   
                -1); 


  @xref
    <f OtxDrvConnectPipe>
    <nl><f OtxDrvCreatePipe>
    <nl><f OtxDrvClose>
    <nl><f OtxDrvConnectLogicalDevice>
    <nl><t OtxPipeEndE>
    <nl><t OTX_HANDLE>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvDisconnectPipe(
  IN OTX_HANDLE       hPipe,            // @parm Pipe to be disconnected
  IN OtxPipeEndE      ePipeEnd          // @parm End of the pipe to be disconnected. See Also <t OtxPipeEndE>
);


/******************************************************************************

    Creating a CT Bus board interconnect device to represent a physical connection
    between two or more boards.

    One entry is added for each CT Bus board interconnect in use, i.e. one for 
    each cable. Then an entry is added for each board, indicating which CT Bus 
    board interconnect is connecting to that board.

    Here is an example:


REGEDIT4

[HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\OtxDrv\Parameters]

[HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\OtxDrv\Parameters\Board]

[HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\OtxDrv\Parameters\Board\T2PX06990036]
"BoardInterConnect"=dword:00000001

[HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\OtxDrv\Parameters\BoardInterConnect]

[HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\OtxDrv\Parameters\BoardInterConnect\0]
"HwyTimeSlots"=dword:00000020

[HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\OtxDrv\Parameters\BoardInterConnect\1]
"HwyTimeSlots"=dword:00000020

[HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\OtxDrv\Parameters\T2PX06990036]
"BoardInterConnect"=dword:00000000


    Note that the HwyTimeSlots must be present to signify the presence of a CT 
    Bus board interconnect cable. The number determines the speed of the 
    interconnect in timeslots per highway. For now only the value 32 is allowed 
    (20 hex).


******************************************************************************/

/******************************************************************************

  @func
    Inform the driver of the physical board connections for a CT Bus. This 
    function should be called once for every board which is connected to 
    this CT Bus.

  @comm
    A CT Bus device is a physical device representing the CT Bus cable
    connection between two or more boards.

    When two ore more boards are connected together with the appropriate CT Bus 
    board interconnect cable it is possible to connect two devices with a pipe
    even if they are located on different board. Before pipes can connect over 
    separate boards the driver must be informed about what types of board 
    interconnects exists, and how the boards are connected. This is done by 
    calling this function.

    The CT Bus device must be opened by calling <f OtxDrvOpenPhysicalDevice> 
    prior to calling this function. The same applies to all the boards which 
    are connected to the CT Bus.

    The CT Bus does not have a direction like a pipe. The direction of each 
    individual timeslot is determined at the time a pipe is connected. A Pipe 
    Device must utilize the CT Bus to reach a different board. The direction 
    of the pipe then determines the direction of that timeslot in the CT Bus.

    The OTX driver supports a CT bus capacity of 2048 kbps which corresponds 
    to 32 timeslots/highway.

  @ex |
    // connect to the Physical CT Buses
    OTX_INT32 nCtBus;
    
    for (nCtBus = 0; nCtBus < 1; nCtBus++) {
      // open OTX_DEVICE_CT_BUS
      result = OtxDrvOpenPhysicalDevice(
                  0, 
                  OTX_DEVICE_CT_BUS, 
                  nCtBus, 
                  nCtBus, 
                  g_hEventQueue, 
                  &g_hCtBus[nCtBus]); 
    }

    for (iBrd = 0; iBrd < 2; iBrd++) {

      // Connect (get handles) to the physical devices on two boards
      DemoOpenPhysicalDevices(iBrd);
    }

    // inform the driver that our two  boards are connected to one CT Bus
    for (iBrd = 0; iBrd < 2; iBrd++) {
      result = OtxDrvConnectBoard(
                  g_hCtBus[iBrd/2], 
                  g_hPhysDevs[iBrd].hThorController); 
    }


  @xref
    <f OtxDrvOpenPhysicalDevice>
    <nl> <e OtxPhysicalDeviceTypeE.OTX_DEVICE_CT_BUS>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvConnectBoard(
  IN OTX_HANDLE       hCtBus,           // @parm Handle to a CT Bus 
  IN OTX_HANDLE       hBoard            // @parm Handle to a board (Nic) device which is connected to this CT bus
);


/******************************************************************************

  @func
    Connect an input of a Logical device to another logical device or to a pipe
    connected to the physical device implementing the logical device.

  @comm
    Two logical devices (which are implemented within one physical device)
    or a logical device and a pipe connected to the physical device implementing
    the logical device can be connected in series with this function. If the 
    devices need to be connected in parallel, call this function twice with 
    the same output device but with two different input devices.

    The connection point index is only needed for devices that have more than
    one inputs of outputs, e.g. a mixer or a conference bridge, and even in that
    case it can be left as -1 if the connection points are 100% symmetrical,
    e.g. a symmetrically mixed conference bridge.

    see example on <f OtxDrvConnectPipe>

  @devnote
    This function can only be used for logical devices and pipes. Physical
    devices must be connected using the function <f OtxDrvConnectPipe>

  @xref
    <f OtxDrvDisconnectLogicalDevice>
    <nl><f OtxDrvConnectPipe>
    <nl><f OtxDrvCreateLogicalDevice>
    <nl><t OTX_HANDLE>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvConnectLogicalDevice(
  IN OTX_HANDLE       hFromDevice,      // @parm Logical Device or Pipe whose output is to be connected
  IN OTX_INT32        iFromConnectPoint,// @parm Output connection point to connect, -1 = don't care
  IN OTX_HANDLE       hToDevice,        // @parm Logical Device or Pipe whose input is to be connected
  IN OTX_INT32        iToConnectPoint   // @parm Connection point on the second device to connect to, -1 = don't care
);


/******************************************************************************

  @func
    Disconnect an input of a Logical device from another logical device or 
    from a pipe connected to the physical device implementing the logical device.

  @comm
    After disconnection, the Logical device may be re-connected to another
    logical device or a pipe with function <f OtxDrvConnectLogicalDevice>

    The connection point index must be the same connection point used when
    OtxDrvConnectLogicalDevice was called.

  @devnote
    This function can only be used for logical devices and pipes. Physical
    devices must be disconnected using the function <f OtxDrvDisconnectPipe>

  @ex |
    // break Cross-Connect 
  
    // disable the logical devices
    result = OtxDrvDisable(
                hDataConverter[0], 
                OTX_TASK_SYNC, 
                5000L); 
    result = OtxDrvDisable(
                hDataConverter[1], 
                OTX_TASK_SYNC, 
                5000L);
  
    // hDataConverter[0])
    result = OtxDrvDisconnectLogicalDevice(
                hLiDspPipe[0],      
                -1, 
                hDataConverter[0], 
                -1); 
    result = OtxDrvDisconnectLogicalDevice(
                hDataConverter[0],  
                -1, 
                hDspLiPipe[1],     
                -1); 
  
    // hDataConverter[1]
    result = OtxDrvDisconnectLogicalDevice(
                hLiDspPipe[1],      
                -1, 
                hDataConverter[1], 
                -1); 
    result = OtxDrvDisconnectLogicalDevice(
                hDataConverter[1],  
                -1, 
                hDspLiPipe[0],     
                -1); 

  @xref
    <f OtxDrvConnectLogicalDevice>
    <nl><f OtxDrvConnectPipe>
    <nl><f OtxDrvCreateLogicalDevice>
    <nl><t OTX_HANDLE>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvDisconnectLogicalDevice(
  IN OTX_HANDLE       hFromDevice,      // @parm Logical Device or Pipe whose output is to be disconnected
  IN OTX_INT32        iFromConnectPoint,// @parm Output connection point to be disconnected, -1 = disconnect all
  IN OTX_HANDLE       hToDevice,        // @parm Logical Device or Pipe whose input is to be disconnected
  IN OTX_INT32        iToConnectPoint   // @parm Connection point on the second device to be disconnect to, -1 = disconnect all
);


/******************************************************************************

  @func
    Create a Notification Event

  @comm
    An Event Handle if registered with an Event Queue can be used by a Win32 
    app to wait efficiently on events to be added to the event queue. That way 
    unnecessary polling with the <f OtxDrvGetEventData> can be avoided.

    OTX_RESULT result;
  
  @ex |
    // create the event queue to be used by the driver to inform the 
    // application about events

    // create a notification Event
    result = OtxDrvCreateEvent(&g_hNotificationEvent); 

    // create Event Queue for driver notifications
    result = OtxDrvCreateEventQueue(
                2048, 
                g_hNotificationEvent, 
                &g_hEventQueue); 

    // enable Event Queue
    result = OtxDrvEnable(
                g_hEventQueue, 
                OTX_TASK_SYNC, 
                5000L); 


  @xref
    <f OtxDrvWaitForSingleEvent>
    <nl><f OtxDrvWaitForMultipleEvents>
    <nl><f OtxDrvCreateEventQueue>
    <nl><f OtxDrvCloseEvent>
    <nl><t OTX_EVENT>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvCreateEvent(
  OUT OTX_EVENT       *phEvent          // @parm Event handle, ready for use. See Also <t OTX_EVENT>
);


/******************************************************************************

  @func
    Set an event into a signaled state.

  @comm
    This function allows a previously created event to be set into signalled
    state. This function can be used by layered applications where lower layer
    module can signal an event to upper layer module which can be waiting
    efficiently using the <f OtxDrvWaitForSingleEvent> function.

  @ex |

    // just a demostration of usage of OtxDrvSignalEvent() and OtxDrvCloseEvent()
    OTX_EVENT g_hEvent;

    void *DemoTestThreadFunc(
      void *lpParameter   // thread data
    )
    {
      int i, j, x;
      // Thread Running, waiting for event
      OtxDrvWaitForSingleEvent(g_hEvent, OTX_TIME_INFINITY);

      // just spend some time
      for (i=0; i<20; i++) {
        for (j=0;j<1000; j++) { 
          x = 3 * i + 4 * j;
        }
      }
      DemoLogf("Thread exiting\n");
    
      return 0;
    }


    if (OTX_SUCCESS(OtxDrvCreateEvent(&g_hEvent))) {
      // starting Thread 
      OsaThreadStart(DemoTestThreadFunc);
      
      OsaThreadSleep(0);
      DemoLogf("Hit return to signaling event\n");
      getch();
      OtxDrvSignalEvent(g_hEvent);
      OsaThreadSleep(0);

      DemoLogf("Hit return to close event and exit\n");
      getch();
      OtxDrvCloseEvent(g_hEvent);

    } else {
      DemoLogf("Failed to create event\n");
      getch();
    }


  @xref
    <nl><f OtxDrvCreateEvent>
    <nl><f OtxDrvWaitForSingleEvent>
    <nl><f OtxDrvWaitForMultipleEvents>
    <nl><f OtxDrvCloseEvent>
    <nl><t OTX_EVENT>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvSignalEvent(
  IN OTX_EVENT        hEvent            // @parm Handle to an Event to be set into signalled state. See Also <t OTX_EVENT>
);


/******************************************************************************

  @func
    Close a Notification Event

  @comm
    Make sure there are no event queues that use event. Closing an event
    that is still in use will crash the system eventually.

    see example on <f OtxDrvSignalEvent>

  @xref
    <f OtxDrvWaitForSingleEvent>
    <nl><f OtxDrvWaitForMultipleEvents>
    <nl><f OtxDrvCreateEventQueue>
    <nl><t OTX_EVENT>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvCloseEvent(
  IN OTX_EVENT        hEvent            // @parm Event handle to close. See Also <t OTX_EVENT>
);


/******************************************************************************

  @func
    Create a Notification Event Queue

  @comm
    An Event Queue is used to serialize asynchronous events detected by the
    driver. Events can be anything that occur spontaneously or as a result
    of a user incurred action. It's the users responsibility to pump event
    messages out of the queue and act on them accordingly. Some messages
    are complete in the sense that they contain all infomation associated
    with the event, whereas other actions require the user to read out more
    info from the object itself (the object handle is always passed in the
    event message).

    see example on <f OtxDrvCreateEvent>

  @xref
    <f OtxDrvCreateEvent>
    <nl><f OtxDrvOpenPhysicalDevice>
    <nl><f OtxDrvCreateLogicalDevice>
    <nl><t OTX_EVENT>
    <nl><t OTX_HANDLE>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvCreateEventQueue(
  IN  OTX_UINT32      nSize,            // @parm Size of the queue to be created in # of messages
  IN  OTX_EVENT       hEvent,           // @parm Event which a Win32 app can
                                        // Wait for events to occur, NULL if
                                        // no notification needed (or if not a Win32 application).
  OUT OTX_HANDLE      *phEventQueue     // @parm Event Queue handle

);


/******************************************************************************

  @func
    Fetch information about events that occurred in the driver.

  @comm
    Data about the events are returned in the order they occured.

    Each event is time stamped.

    Win32 applications can wait efficiently for events to occur by setting
    up a Win32 Event with the driver, see <f OtxWaitForSingleEvent>

    see example on <f OtxDrvWaitForSingleEvent>

  @xref
    <f OtxDrvCreateEventQueue>
    <nl><f OtxDrvGetEventDataEx>
    <nl><f OtxDrvEnable>
    <nl><f OtxDrvDisable>
    <nl><f OtxDrvReset>
    <nl><t OtxEventDataS>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvGetEventData(
  IN  OTX_HANDLE      hEventQueue,      // @parm Event queue to read event from
  OUT OtxEventDataS   *pEventData       // @parm Pointer to place-holder for event data
                                        //       See also <t OtxEventDataS>
);


/******************************************************************************

  @func
    Fetch multiple events that occurred in the driver. This function has the same
    functionality as the <f OtxDrvGetEventData> function, except that this function
    is able to fetch multiple event from the driver in one call (and thus reduce 
    overhead introduce by context-switching).

  @ex |

    void  *DemoIdleFunction(
      void *lpParameter   // thread data
    )
    {
      OTX_RESULT result;

      while (1) {
        OtxDrvWaitForSingleEvent(
           g_hNotificationEvent, 
           OTX_TIME_INFINITY);

        do {
          // Multiple reads of events
          OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
          OTX_INT32 coEventsFetched;
          OTX_INT32 i;

          result = OtxDrvGetEventDataEx(
                      g_hEventQueue, 
                      sqEventData, 
                      10, 
                      &coEventsFetched); 

          if (OTX_RESULT_CODE(result) == OTX_S_OK) {
            // We have a driver event (one at least)
            //DemoLogf("OtxDrvGetEventDataEx() fetched %d events\n", coEventsFetched);

            assert(coEventsFetched <= coEventsFetched);

            for (i = 0; i < coEventsFetched; i++) {
              // Print the event data.
              // In a real application branch based on the data
              DemoPrintEvent(&sqEventData[i]);
            }

          }
        } while (OTX_RESULT_CODE(result) == OTX_S_OK);

      }

      return 0;

    }

  @xref
    <f OtxDrvGetEventData>    
    <nl><t OtxEventDataS>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvGetEventDataEx(
  IN  OTX_HANDLE      hEventQueue,        // @parm Event queue to read event from
  OUT OtxEventDataS   *psqEventData,      // @parm Pointer to place-holder array for event data (<t OtxEventDataS>)
  IN OTX_INT32        coEventData,        // @parm Capacity of <p psqEventData> (in number of <t OtxEventDataS> elements), i.e. maximum number
                                          //       of events which the driver can entry in <p psqEventData>.
  OUT OTX_INT32       *pcoEventsReturned  // @parm Number of events returned by this function (number of elements which the driver entered in <p psqEventData>).
);


/******************************************************************************

  @func
    Place an event data to an event queue.

  @comm
    This function allows a module to place data into a previously created
    event queue. This function can be used by layered applications where lower
    layer module can pass an event to an upper layer module which can be
    waiting efficiently using the <f OtxDrvWaitForSingleEvent> function.

    The driver will time stamp the event.

  @ex |

    // set an event in the event queue
    OtxDrvSetEventDataEx(
       g_hEventQueue, 
       OTX_DEVICE_LI_T1E1, 
       0, 
       g_hPhysDevs.hLi[nCurrentLi], 
       OTX_T1E1_EC_AIS, 
       0, 
       9,
       nTimeStamp);

  @xref
    <f OtxDrvCreateEventQueue>
    <nl><f OtxDrvGetEventData>
    <nl><t OtxEventDataS>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvSetEventDataEx(
  IN  OTX_HANDLE       hEventQueue,     // @parm Event queue to be used
  IN  OTX_UINT32       eDeviceType,     // @parm Device Type with which the event is associated
  IN  OTX_UINT32       nSourceId,       // @parm Device Id of the device that the event is associated with (supplied by the user when the device for opened/created)
  IN  OTX_HANDLE       hDevice,         // @parm Handle to the Device with which the event is associated
  IN  OTX_NOTIFY_CAUSE nCode,           // @parm Device specific cause code indicating what happened. Each device types has it own set of event cause codes
  IN  OTX_TASK_REF     nRequestId,      // @parm If the message is for a previous user request this member holds the user assigned request ID
  IN  OTX_UINT32       nParam,          // @parm Optional parameter with user defined meaning
  IN  OTX_DATETIME     nTimeStamp       // @parm Timestamp to mark the event with
);

/******************************************************************************

  @func
    Place an event data to an event queue.

  @comm
    This function allows a module to place data into a previously created
    event queue. This function can be used by layered applications where lower
    layer module can pass an event to an upper layer module which can be
    waiting efficiently using the <f OtxDrvWaitForSingleEvent> function.

    The driver will time stamp the event.

  @ex |

    // set an event in the event queue
    OtxDrvSetEventData(
       g_hEventQueue, 
       OTX_DEVICE_LI_T1E1, 
       0, 
       g_hPhysDevs.hLi[nCurrentLi], 
       OTX_T1E1_EC_AIS, 
       0, 
       9);

  @xref
    <f OtxDrvCreateEventQueue>
    <nl><f OtxDrvGetEventData>
    <nl><t OtxEventDataS>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvSetEventData(
  IN  OTX_HANDLE       hEventQueue,     // @parm Event queue to be used
  IN  OTX_UINT32       eDeviceType,     // @parm Device Type with which the event is associated
  IN  OTX_UINT32       nSourceId,       // @parm Device Id of the device that the event is associated with (supplied by the user when the device for opened/created)
  IN  OTX_HANDLE       hDevice,         // @parm Handle to the Device with which the event is associated
  IN  OTX_NOTIFY_CAUSE nCode,           // @parm Device specific cause code indicating what happened. Each device types has it own set of event cause codes
  IN  OTX_TASK_REF     nRequestId,      // @parm If the message is for a previous user request this member holds the user assigned request ID
  IN  OTX_UINT32       nParam           // @parm Optional parameter with user defined meaning
);


/******************************************************************************

  @func
    Wait for a notification event from the driver

  @comm
    Waits efficiently using a Win32 event.

    Specify nMilliseconds = <t OTX_TIME_INFINITY> to wait forever.

  @rdesc Returns one of the following values:

    @flag OTX_E_HANDLE | Handle invalid
    @flag OTX_S_TIMEOUT | Wait timed out
    @flag OTX_S_OK | Event occured
    @flag OTX_E_FAIL | Non-specified failure

  @ex |
    // check for Events from the OTX Driver
    while (1) {
      OtxDrvWaitForSingleEvent(
         g_hNotificationEvent, 
         OTX_TIME_INFINITY);
      do {
        result = OtxDrvGetEventData(
                    g_hEventQueue, 
                    &eventData); 

        if (OTX_RESULT_CODE(result) == OTX_S_OK) {
          // We have a driver event
          // Print the event data.
          // In a real application branch based on the data
          DemoPrintEvent(&eventData);
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  

  @xref
    <f OtxDrvWaitForMultipleEvents>
    <nl><f OtxDrvCreateEvent>
    <nl><t OTX_EVENT>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvWaitForSingleEvent(
  IN OTX_EVENT        hEvent,           // @parm Event handle, initilized with <f OtxDrvCreateEvent>
  IN OTX_TIME         nWaitMaxMs        // @parm Max time to wait in millisconds
);


/******************************************************************************

  @func
    Wait for a notification event from the driver

  @comm
    Waits efficiently using a Win32 event.

    Specify nMilliseconds = OTX_TIME_INFINITY to wait forever.

  @xref
    <f OtxDrvWaitForSingleEvent>
    <nl><f OtxDrvCreateEvent>
    <nl><t OTX_EVENT>
    <nl><t OTX_BOOL>

  @todo

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvWaitForMultipleEvents(
  IN  OTX_INT32       coEventHandles,   // @parm Event handle count
  IN  OTX_EVENT       sqhEvents[],      // @parm Array of event handles. See Also <t OTX_EVENT>
  IN  OTX_TIME        nWaitMaxMs,       // @parm Max time to wait in millisconds
  IN  OTX_BOOL        bWaitAll,         // @parm Specifies the wait type. If OTX_TRUE, the function returns
                                        //       when the state all objects in the lpHandles array is signaled.
                                        //       If OTX_FALSE, the function returns when the state of any
                                        //       one of the objects set to is signaled. In the latter case,
                                        //       the return value indicates the object whose state caused the
                                        //       function to return.  See also <t OTX_BOOL>
  OUT OTX_INT32       *pnEventFired     // @parm index to event in array that fired
);


/******************************************************************************

  @func
    Cancels a pending task for a particular device.

  @comm
    If a device has several tasks with the same reference number, the oldest
    task is cancelled. Only one or zero tasks are removed for each call.

  @xref
    <t OTX_TASK_REF>

  @todo

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvCancelTask(
  IN OTX_HANDLE       hDevice,          // @parm Handle to the device
  IN OTX_TASK_REF     nTaskRef          // @parm Callers task reference, as given earlier. See <t OTX_TASK_REF>
);


#if 0
/******************************************************************************

  @func
    Transparent device communication.

  @comm
    Mostly used to communicate with firmware modules and firmware devices. 
    Since the capabilities of the firmware are transparent to the driver, 
    this is the only way firmware modules and devices can be controlled by an 
    application using the OtxDrv API.
  
  @xref
    <t OTX_HANDLE>
    <nl><t OTX_TASK_REF>
    <nl><t OTX_TIME>

  @todo

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvDeviceControl(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to device
  IN  OTX_INT32       nCommandCode,     // @parm Code indicating requested operation
  IN  OTX_UINT8       *pInBuffer,       // @parm Input parameters in serial format
  IN  OTX_UINT32      cbInBuffer,       // @parm Byte count of input parameters
  OUT OTX_UINT8       *pOutBuffer,      // @parm Buffer to receive output parameters in serial format
  IN  OTX_UINT32      cbOutBuffer,      // @parm Size in bytes of output buffer. Must be large enough for the given command
  OUT OTX_UINT32      cbReturned,       // @parm Actual number of bytes returned (may differ from cbOutBuffer).
  IN  OTX_TASK_REF    nTaskRef,         // @parm See <t OTX_TASK_REF>
  IN  OTX_TIME        nWaitMaxMs        // @parm See <t OTX_TIME>
);
#endif

/******************************************************************************

  @func
    Polls for Events from the driver.

  @comm
    This function can be used if interrupts can not be used for some reason.
    Instead of an interrupt service routine servicing interrupts upon arrival,
    the application can constantly poll this function and force the interrupt
    service routine to be executed periodically without having interrupts
    enabled.

  @ex |
    OtxEventDataS eventData;
    OTX_RESULT result;

    OtxDrvPollEvents(g_hPhysDevs.hThorController0);

    result = OtxDrvGetEventData(g_hEventQueue, &eventData);

    if (OTX_RESULT_CODE(result) == OTX_S_OK) {
      // We have a driver event
      // Printf the event data.
      // In a real application branch based on the data
      DemoPrintEvent(&eventData);
    }  


******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvPollEvents(
  IN  OTX_HANDLE      hDevice           // @parm Handle to device
);


/******************************************************************************
  @func
    Initializes an iterator so that it can be used to retrieve data for the
    first element/device in a sequence

  @comm
    see example on <f OtxDrvEnumDevice>  

  @xref
    <f OtxDrvEnumDeviceType>
    <nl><f OtxDrvEnumDevice>
    <nl><f OtxDrvIsIteratorEnd>
    <nl><t OTX_ITERATOR>


******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvIteratorBegin(
  OTX_ITERATOR        *pItr           // @parm pointer to iterator to initialize. See <t OTX_ITERATOR>
);


/******************************************************************************
  @func
    Checks an iterator to see if it has reached past the end of the sequence
    being enumerated.

  @xref
    <f OtxDrvIteratorBegin>
    <nl><f OtxDrvIteratorNext>
    <nl><t OTX_ITERATOR>

  @todo


******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvIsIteratorEnd(
  OTX_ITERATOR        *pItr           // @parm pointer to iterator to check. See <t OTX_ITERATOR>
);


/******************************************************************************
  @func
    Enumerates over all device types supported by a parent device

  @comm
    Retrievs data for item referenced by the iterator

    The iterator is incremented to reference the next item in the sequence. 
    It can be used in a subsequent call to the same function to fetch data for 
    the next element in the same sequence. Repeated calls to the same function 
    with the same iterator eventually fetches data for all items in this sequence.

    The iterator must already be initialized with a call to <f OtxDrvIteratorBegin>
    before the first call to this function.

    It is important that the iterator is used on the same sequence of items 
    between each call to <f OtxDrvIteratorBegin>, i.e. the same function must 
    be called with the same handle and the same iterator.

    However, it is possible for multiple iterators to be used on the same sequence
    independently, at the same time, even in separate threads.

    see example on <f OtxDrvEnumDevice> 

  @xref
    <f OtxDrvEnumDevice>
    <nl><f OtxDrvIteratorBegin>
    <nl><f OtxDrvIsIteratorEnd>
    <nl><t OTX_ITERATOR>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvEnumDeviceType(
  IN     OTX_HANDLE   hDevice,          // @parm Handle to the physical parent device
  IN OUT OTX_ITERATOR *pItr,            // @parm Enumeration start position. See <t OTX_ITERATOR>
  OUT    OTX_INT32    *peDeviceType     // @parm Physical or logical device type. See <t OtxPhysicalDeviceTypeE> and <t OtxLogicalDeviceTypeE>
);


/******************************************************************************
  @func
    Enumerates over all device numbers for a specific type of devices supported
    by a parent device.

  @comm
    Retrievs data for item referenced by the iterator

    After the call, the iterator has been incremented to reference the next 
    item in the sequence. It can be used in a subsequent call to the same 
    function to fetch data for the next element in the same sequence. Repeated 
    calls to the same function with the same iterator eventually fetches data 
    for all items in this sequence. Note that it is not possible to increment 
    the iterator in any other way.

    The iterator must already be initialized with a call to <f OtxDrvIteratorBegin>
    before the first call to this function.

    It is important that the iterator is used on the same sequence of items 
    between each call to <f OtxDrvIteratorBegin>, i.e. the same function must 
    be called with the same handle value (hDevice) and the same iterator (pItr).

    However, it is possible for multiple iterators to be used on the same sequence
    independently, at the same time, even in separate threads.

  @ex |
    {
      OTX_UINT32   eDevType;
      OTX_ITERATOR itrChildTypes;
      OTX_ITERATOR itrChildrenOfSameType;
      OTX_HANDLE hParentDevice = 0;

      // Initialize the opaque iterator for the hosted types
      OtxDrvIteratorBegin(&itrChildTypes);
      
      while (OTX_SUCCESS(OtxDrvEnumDeviceType(hParentDevice, &itrChildTypes, (OTX_INT32 *)&eDevType))) {   // Note that we must use a zero handle to get the children of the root level
      
        OTX_HANDLE hChildDevice;
        // Initialize the opaque iterator for the devices within the same type
        OtxDrvIteratorBegin(&itrChildrenOfSameType);
        
        while (OTX_SUCCESS(OtxDrvEnumDevice(hParentDevice, eDevType, &itrChildrenOfSameType, &hChildDevice))) {
          if (hChildDevice != OTX_INVALID_HANDLE_VALUE) {
            OTX_UINT32 eDevType2;
            OTX_HANDLE hParentDevice2;

				    // Here we have a handle for a child device, get the attributes
            DemoLogf("%s_%d (%s, handle = %#x)\n", 
            DemoGetDeviceTypeString(hChildDevice), 
            DemoGetDeviceNo(hChildDevice), 
            (eDevType >= 0) ? (eDevType == 0 ? "UnKnown" : "Physical") : "Logical", hChildDevice);

            // Print device specific attributes
            switch (eDevType) {
              case OTX_DEVICE_THOR_PCI:
              case OTX_DEVICE_THOR_8_PCI:
              {
           	    OTX_RESULT      result;

                // Read serial number from the board
                ReadSerialNum( hParentDevice, hChildDevice, eDevType );

                if (eDevType) {
                  DemoLogf("Opening board# %d using OtxDrvOpenPhysicalDeviceEx()\n", DemoGetDeviceNo(hChildDevice));
              
                  result = OtxDrvOpenPhysicalDeviceEx(
                              hChildDevice, 
                              DemoGetDeviceNo(hChildDevice), 
                              OTX_INVALID_HANDLE_VALUE); 
                }
              }
              break;
            }

            // Now do the same thing for all children of this device (recursion)
            ms_nIndentLevel += 1;
            EnumerateDevices(hChildDevice);
            ms_nIndentLevel -= 1;
          }
        } // while loop
      } // while loop
    }

  @xref
    <f OtxDrvEnumDeviceType>
    <nl><t OtxDrvIteratorBegin>
    <nl><t OtxDrvIsIteratorEnd>
    <nl><t OTX_ITERATOR>


******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvEnumDevice(
  IN     OTX_HANDLE   hDevice,          // @parm Handle to the physical parent device
  IN     OTX_INT32    eDeviceType,      // @parm Physical or logical device type. See <t OtxPhysicalDeviceTypeE> and <t OtxLogicalDeviceTypeE>
  IN OUT OTX_ITERATOR *pItr,            // @parm Enumeration start position. See <t OTX_ITERATOR>
  OUT    OTX_HANDLE   *phChildDevice    // @parm Handle to device referenced by iterator. This handle can be always be used to query the device, but the device must have been opened for most other operations to succeed
);


/******************************************************************************
  @func
    Translate an Event message into English text describing the event.

  @comm
    The event message string is copied into a user supplied buffer

    see example on <f OtxHdlcReadData>

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvEventCode2String(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device who generated the event
  IN  OTX_INT32       nEventCode,       // @parm Event Cause Code to be translated. See also <t OtxEventDataS>
  IN  OTX_UINT32      nBufLen,          // @parm Length of the User supplied String Buffer
  OUT char            szStrBuf[]        // @parm User supplied String buffer
);


/******************************************************************************
  @func
    Translate a result code into English text describing the code.

  @comm
    The result message string is copied into a user supplied buffer

    Usaully it is used in DemoCheckRc() on demo programs

  @ex |
    
    static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
    {
      if ( OTX_FAILURE(result) ) {

        if (hDevice == OTX_INVALID_HANDLE_VALUE) {
          DemoLogf("ERROR(%#lX) \n", result);
        } else {
          OTX_RESULT otxRc;
          char msgStr[256];
          OtxAttribTypeE  eAttribType;   
          OTX_UINT32      cbAttribValue; 
          OTX_BOOL        bReadOnly;     
          OTX_UINT32      eDeviceType = 999;    
          OTX_UINT32      nDeviceNumber = 999;  
        
          // Get the Device Type for hHandle
          otxRc = OtxDrvGetAttributeType(
                     hDevice, 
                     OTX_ATTR_DEVICE_TYPE, 
                     &eAttribType, 
                     &cbAttribValue, 
                     &bReadOnly); 

          if (OTX_SUCCESS(otxRc)) {
            assert(eAttribType == OTX_ATTRIB_UINT32);
            assert(cbAttribValue == sizeof(OTX_UINT32));
            OtxDrvGetAttributeValueUint32(
                     hDevice, 
                     OTX_ATTR_DEVICE_TYPE, 
                     &eDeviceType); 
          } 
        
          // Get the Device Number for hHandle
          otxRc = OtxDrvGetAttributeType(
                     hDevice, 
                     OTX_ATTR_DEVICE_NO, 
                     &eAttribType, 
                     &cbAttribValue, 
                     &bReadOnly); 

          if (OTX_SUCCESS(otxRc)) {
            assert(eAttribType == OTX_ATTRIB_UINT32);
            assert(cbAttribValue == sizeof(OTX_UINT32));
            OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
          }
        
          // Convert the result code to a more descriptive string
          otxRc = OtxDrvResultCode2String(
                     hDevice, 
                     result, 
                     256, 
                     &msgStr[0]);

          if (OTX_FAILURE(otxRc)) {
            strcpy(msgStr, "Cannot translate this result code.");
          }
          DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
        }

        DemoExitQuestion();
      }
    }


******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvResultCode2String(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device who generated the result code
  IN  OTX_RESULT      nResultCode,      // @parm Result Code to be translated. See also <t OTX_RESULT>
  IN  OTX_UINT32      nBufLen,          // @parm Length of the User supplied String Buffer
  OUT char            szStrBuf[]        // @parm User supplied String buffer
);


/******************************************************************************
  @func
    Read a 32-bit register from a physical device

  @todo
******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvReadReg32(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device to read from
  IN  OTX_UINT32      nAddress,         // @parm Register address within the device
  OUT OTX_UINT32      *pnValue          // @parm 32-bit value read from the device
);                                      


/******************************************************************************
  @func
    Read a 16-bit register from a physical device

  @ex | 

    OTX_UINT32 readVal;
    char strBuf[12];
    static OTX_UINT32 nAddress = 0;

    // read from a 16-bit register in board device
    DemoLogf("Which Address (0x%X): ", nAddress);
    OsaConsoleGetStr(strBuf); 
    if (*strBuf) {
      nAddress = strtoul(strBuf, NULL, 0);
    }

    result = OtxDrvReadReg16(
                g_hPhysDevs.hBoardController, 
                nAddress, 
                &readVal); 
    
    // display
    DemoLogf("Brd Address 0x%X = %.4X\n", nAddress, (OTX_UINT16)readVal);

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvReadReg16(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device to read from
  IN  OTX_UINT32      nAddress,         // @parm Register address within the device
  OUT OTX_UINT32      *pnValue          // @parm 16-bit value read from the device
);                                      


/******************************************************************************
  @func
    Read an 8-bit register from a physical device

  @ex |

    OTX_UINT32 readVal;
    char strBuf[12];
    static OTX_UINT32 nAddress = 0;

    // read from an 8-bit register in current Li
    DemoLogf("Which Address (0x%X): ", nAddress);

    OsaConsoleGetStr(strBuf); 
    if (*strBuf) {
      nAddress = strtoul(strBuf, NULL, 0);
    }

    result = OtxDrvReadReg8(
                g_hPhysDevs.hLi[nCurrentLi], 
                nAddress, 
                &readVal); 

    // display
    DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvReadReg8(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device to read from
  IN  OTX_UINT32      nAddress,         // @parm Register address within the device
  OUT OTX_UINT32      *pnValue          // @parm 8-bit value read from the device
);                                      


/******************************************************************************
  @func
    Write a 32-bit value to a 32-bit register in a physical device

  @todo
******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvWriteReg32(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device to write to
  IN  OTX_UINT32      nAddress,         // @parm Register address within the device
  OUT OTX_UINT32      nValue            // @parm 32-bit value to be written to the device
);                                      


/******************************************************************************
  @func
    Write a 16-bit value to a 16-bit register in a physical device

  @ex |

    char strBuf[12];
    static OTX_UINT32 nAddress = 0;
    static OTX_UINT32 nValue = 0;

    // write to a 16-bit register in board device
    DemoLogf("Which Address (0x%X): ", nAddress);
    OsaConsoleGetStr(strBuf);   
    if (*strBuf) {
      nAddress = strtoul(strBuf, NULL, 0);
    }
    DemoLogf("Which Value (0x%X): ", nValue);
    OsaConsoleGetStr(strBuf);   
    if (*strBuf) {
      nValue = strtoul(strBuf, NULL, 0);
    }

    result = OtxDrvWriteReg16(
                g_hPhysDevs.hBoardController, 
                nAddress, 
                nValue); 

    DemoLogf("Wrote %.4X to Brd Address 0x%X\n", (OTX_UINT16)nValue, nAddress);

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvWriteReg16(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device to write to
  IN  OTX_UINT32      nAddress,         // @parm Register address within the device
  OUT OTX_UINT32      nValue            // @parm 16-bit value to be written to the device
);                                      


/******************************************************************************
  @func
    Write an 8-bits value to an 8-bit register in a physical device

  @ex |

    char strBuf[12];
    static OTX_UINT32 nAddress = 0;
    static OTX_UINT32 nValue = 0;

    // write to an 8-bit register in current Li
    DemoLogf("Which Address (0x%X): ", nAddress);
    OsaConsoleGetStr(strBuf);  
    if (*strBuf) {
      nAddress = strtoul(strBuf, NULL, 0);
    }
    DemoLogf("Which Value (0x%X): ", nValue);
    OsaConsoleGetStr(strBuf);   
    if (*strBuf) {
      nValue = strtoul(strBuf, NULL, 0);
    }

    result = OtxDrvWriteReg8(
                g_hPhysDevs.hLi[nCurrentLi], 
                nAddress, 
                nValue); 

    DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvWriteReg8(
  IN  OTX_HANDLE      hDevice,          // @parm Handle to the device to write to
  IN  OTX_UINT32      nAddress,         // @parm Register address within the device
  OUT OTX_UINT32      nValue            // @parm 8-bit value to be written to the device
);                                      


/******************************************************************************
  // Debug function

******************************************************************************/
OTX_EXPORT OTX_RESULT OTX_API OtxDrvDbgDump(
  IN  OTX_HANDLE      hDevice,
  IN  OTX_UINT32      nCmd
);                    



#ifdef __cplusplus
}  // extern "C"
#endif

/******************************************************************************

  Copyright (C) Odin TeleSystems Inc., 1998 - 2003


  All rights reserved. No part of this computer program may be reproduced,
  stored in any retrieval system, or transmitted, in any form or by any
  means, electronic, mechanical, photocopying, recording, or otherwise,
  without the prior written permission of Odin TeleSystems Inc., or without
  a permission stipulated in the agreement/contract under which this computer
  program was supplied.

******************************************************************************/
#endif  // OTX_DRV_H
